# -*- coding:utf-8 -*-

from lawes.db.models.query import QuerySet
from lawes.utils import six
from lawes.db.models.options import Options

class ModelBase(type):

    def __new__(cls, name, bases, attrs):
        """  changed Field to true type
        """
        super_new = super(ModelBase, cls).__new__
        # if it is not the ModelBase , return
        parents = [b for b in bases if isinstance(b, ModelBase)]
        if not parents:
            return super_new(cls, name, bases, attrs)

        # Create the class.
        # new_class = super_new(cls, name, bases, attrs)
        # module = attrs.pop('__module__')
        new_class = super_new(cls, name, bases, attrs)
        meta = attrs.pop('Meta', None)
        if not meta:
            meta = getattr(new_class, 'Meta', None)
        # db_table is the collection name in mongodb
        db_table = new_class.__module__.split('.')[-1] + '_' + new_class.__name__.lower()
        kwargs = {"db_table": db_table}
        new_class._meta = Options(meta, **kwargs)

        # set the local_fields at Options: {'name': lawes.db.models.fields.CharField}
        for obj_name, obj in attrs.items():
            if hasattr(obj, 'contribute_to_class'):
                new_class._meta.add_field(obj_name=obj_name, obj=obj)

        # create the objects
        new_class.objects = QuerySet(model=new_class)

        return new_class


class Model(six.with_metaclass(ModelBase)):

    pk_attname = '_id'
    save_fields = []


    def __init__(self, *args, **kwargs):
        # set the real value to the model
        for obj_name in self._meta.local_fields:
            obj = self._meta.local_fields[obj_name]
            setattr(self, obj_name, obj.value)

        # creat the index and unique in db
        # self.objects.init_index(db_indexs = self._meta.db_indexs)

        super(Model, self).__init__()


    def __setattr__(self, key, value):
        super(Model, self).__setattr__(key, value)
        if hasattr(self, '_id'):
            self.save_fields.append(key)


    def save(self):
        self._save_table()


    def _save_table(self):

        pk_val = self._get_pk_val()
        # pk_set is true for updating, pk_set is true for inserting
        pk_set = pk_val is not None
        data = self.to_dict(fields='save_fields') if pk_set else self.to_dict()
        # the format must be correct
        checking_keys = set(self._meta.local_fields) & set(data)
        for key in checking_keys:
            self._meta.local_fields[key].check_type(value=data[key])
        if pk_set:
            self._do_update(data=data)
        else:
            result = self._do_insert(data=data)
            setattr(self, self.pk_attname, result)


    def _get_pk_val(self):
        """ get _id: None: INSERT; not None: UPDATE
        """
        if not hasattr(self, self.pk_attname):
            return None
        else:
            return getattr(self, self.pk_attname)


    @classmethod
    def _do_insert(cls, data):
        """ 向 mongodb 插入数据
        """
        return cls.objects._insert(data=data)


    @classmethod
    def _do_update(cls, data):
        """ doing update in mongodb
        """
        return cls.objects._update(data=data)


    def to_dict(self, fields=''):
        """ if the values of the fields is 'save_fields' , only change the fields parts
        """
        if fields == 'save_fields':
            fields_type = self.save_fields
        else:
            fields_type = self._meta.local_fields
        result = {}

        for field in fields_type:
            if hasattr(self, field):
                value = getattr(self, field)
                result[field] = value

        if hasattr(self, '_id'):
            result['_id'] = self._id

        return result


    def to_obj(self, data={}):

        for field in data:
            setattr(self, field, data[field])
        for field in self._meta.local_fields:
            if field in data:
                value = data[field]
            else:
                value = self._meta.local_fields[field].value
            setattr(self, field, value)
        if '_id' in data:
            self._id = data['_id']
        return self


    def delete(self):
        return self.objects._remove(_id=self._id)



# -*- coding:utf-8 -*-

from lawes.db.models.base import Model
from lawes.db.models.fields import *
from lawes.db.models.query import configqueryset

def setup(conf):
    configqueryset._setup(conf=conf)


# -*- coding:utf-8 -*-

from setuptools import setup
from setuptools import find_packages

VERSION = '1.9.2.1'

setup(
    name='Lawes',
    description='',
    long_description='',
    classifiers=[],
    keywords='',
    author='Lawes',
    author_email='haiou_chen@sina.cn',
    url='https://github.com/MrLawes/Lawes',
    license='MIT',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=True,
    install_requires=[
        'pymongo>=2.8',
    ],
    version=VERSION,
)



# -*- coding:utf-8 -*-

from pymongo import MongoClient
from lawes.core.exceptions import MultipleObjectsReturned
from lawes.core.exceptions import DoesNotExist
from lawes.core.exceptions import UniqueError
from lawes.core.exceptions import MongoClientError
from pymongo.errors import OperationFailure
from bson.objectid import ObjectId

CONF_RAESE = """

The correct formal is:
from lawes.db import models
models.setup(conf={'mongo_uri': 'mongodb://127.0.0.1:27017/test', 'db_name': 'testindex'})
self._mongo: %s , self._db: %s
"""

class ConfigQuerySet(object):


    def __init__(self):
        self.mongo = None
        self.db_name = ''


    def _setup(self, conf):
        """ 设置mongodb的连接方式
        """
        if self.mongo:
            return
        if self.db_name:
            return
        if not 'db_name' in conf:
            raise MongoClientError(CONF_RAESE % ('', ''))
        if not 'mongo_uri' in conf:
            raise MongoClientError(CONF_RAESE % ('', ''))
        self.mongo = MongoClient(conf['mongo_uri'])
        self.db_name = conf['db_name'].lower()

# init the MongoClient
configqueryset = ConfigQuerySet()


class QuerySet(object):


    def __init__(self, model=None):
        self.model = model
        self._mongo = configqueryset.mongo
        self._db = configqueryset.db_name        # the name of the db
        self.db_table = model._meta.db_table      # the name of the collection
        if not self._mongo or not self._db:
            raise MongoClientError(CONF_RAESE % (str(self._mongo), str(self._db)))
        self._collection = getattr(self._mongo[self._db], self.db_table)
        self.filter_query = {}                      # using for Model.objects.filter(filter_query)
        self.order_by_query = ()                    # using for Model.objects.order_by(filter_query)
        self.skip = None                            # using for Model.objects.skip(skip)
        self.limit = None                           # using for Model.objects.limit(limit)


    def __iter__(self):
        for data in self._fetch_all():
            yield data


    def __getitem__(self, item):

        if isinstance(item, int):
            self.limit = item
        elif isinstance(item, slice):
            self.skip = item.start
            self.limit = item.stop - item.start
        return self.__iter__()


    def filter(self, **query):
        if query == {}:
            self.filter_query = {}
        else:
            self.filter_query.update(query)
        return self


    def _fetch_all(self, originally=False):
        """ run the sql actually
        :return:
        """
        multi_data = self._collection.find(self.filter_query)
        # order by query
        if self.order_by_query:
            multi_data = multi_data.sort(*self.order_by_query)

        if not self.skip is None:
            multi_data = multi_data.skip(self.skip)
        if not self.limit is None:
            multi_data = multi_data.limit(self.limit)

        if originally is True:
            for data in multi_data:
                yield data
        else:
            for data in multi_data:
                obj = self.model()
                obj = obj.to_obj(data=data)
                yield obj


    def _insert(self, data):
        """
        Inserts a new record for the given model. This provides an interface to
        the InsertQuery class and is how Model.save() is implemented.
        """
        return self._collection.insert(data)


    def _update(self, data):
        """
        Inserts a new record for the given model. This provides an interface to
        the InsertQuery class and is how Model.save() is implemented.
        """
        mongodb_id = data.pop('_id')
        return self._collection.update({'_id': mongodb_id}, {'$set': data}, upsert=True)


    def init_index(self):
        """  create the index_1
        :param db_indexs: {'name': {'unique': False}}
        :return:
        """
        db_indexs = self.model._meta.db_indexs
        try:
            old_index = self._collection.index_information()
        except OperationFailure:
            old_index = {}

        for attr in db_indexs:
            unique = db_indexs[attr].get('unique', False)
            if not attr + '_1' in old_index:
                if unique is True:
                    self._collection.ensure_index(attr, unique=True)
                else:
                    self._collection.ensure_index(attr)
            elif unique is True:
                if not 'unique' in old_index[attr + '_1']:
                    self._collection.ensure_index(attr, unique=True)


    def get(self, *args, **kwargs):
        """
        Performs the query and returns a single object matching the given
        keyword arguments.
        """
        self.filter_query.update(kwargs)
        data = self._collection.find(self.filter_query)
        num = data.count()
        if num == 1:
            obj = self.model()
            obj = obj.to_obj(data=data[0])
            return obj
        if not num:
            raise DoesNotExist(
                "%s matching query does not exist." %
                self.filter_query
            )
        raise MultipleObjectsReturned(
            "findone  %s returned more than one -- it returned %s!" %
            (self.filter_query, num)
        )


    def get_or_create(self, **kwargs):
        """
        :param kwargs:  name='x'
        :return:  a tuple of (object, created), where created is a boolean
        specifying whether an object was created.
        """
        created = False
        try:
            index_information = self._collection.index_information()
        except OperationFailure:
            index_information = {}

        kwargs_keys = set(kwargs.keys())
        index_keys = [ filed for filed in index_information if index_information[filed].get('unique', False)]
        index_keys = [ index_key.replace('_1', '').replace('_-1', '') for index_key in index_keys ]
        need_index_set = kwargs_keys - set(index_keys)
        if need_index_set:
            raise UniqueError('UNIQUE constraint failed: %s: %s, please do collection.ensure_index' % (
            self.db_table, str(need_index_set)))
        self.filter_query.update(kwargs)
        data = self._collection.find(self.filter_query)
        num = data.count()
        if num == 1:
            to_obj_data = data[0]
        elif not num:
            to_obj_data = kwargs
            created = True
        else:
            raise MultipleObjectsReturned(
                "findone  %s returned more than one -- it returned %s!" %
                (self.filter_query, num)
            )

        obj = self.model()
        obj = obj.to_obj(data=to_obj_data)
        if not num:
            obj.save()
        return obj, created


    def order_by(self, *field_names):
        field_names = field_names[0]
        if '-' in field_names:
            field_names = field_names.replace('-', '')
            order_index = -1
        else:
            order_index = 1
        self.order_by_query = (field_names, order_index)
        return self


    def delete(self):
        """
        Deletes the records in the current QuerySet.
        """
        for data in self._fetch_all(originally=True):
            self._remove(_id=str(data['_id']))


    def _remove(self, _id):
        remove_dict = {'_id': ObjectId(_id)}
        self._collection.remove(remove_dict)





from lawes.utils.encoding import python_2_unicode_compatible
from lawes.db.models.fields import Field

DEFAULT_NAMES = ('verbose_name', 'verbose_name_plural', 'db_table', 'ordering',
                 'unique_together', 'permissions', 'get_latest_by',
                 'order_with_respect_to', 'app_label', 'db_tablespace',
                 'abstract', 'managed', 'proxy', 'swappable', 'auto_created',
                 'index_together', 'apps', 'default_permissions',
                 'select_on_save', 'default_related_name')

@python_2_unicode_compatible
class Options(object):
    """ for the model to set default
    """
    def __init__(self, meta, db_table=None):

        self.db_table = db_table
        self.local_fields = {} # {'name': lawes.db.models.fields.CharField }
        self.db_indexs = {}
        self.meta = meta
        # set attribute for Option from Model's Meta
        if meta:
            meta_attrs = self.meta.__dict__.copy()
        else:
            meta_attrs = {}
        for attr_name in meta_attrs:
            if attr_name.startswith('_'):
                continue
            if attr_name in DEFAULT_NAMES:
                setattr(self, attr_name, meta_attrs[attr_name])

    def add_field(self, obj_name, obj):
        self.local_fields[obj_name] = obj
        if isinstance(obj, Field):
            if obj.db_index is True or obj.unique is True:
                self.db_indexs[obj_name] = { 'unique': obj.unique }


    def __str__(self):
        return ''



# -*- encoding: utf-8 -*-

from lawes.utils import six

def python_2_unicode_compatible(klass):
    """
    A decorator that defines __unicode__ and __str__ methods under Python 2.
    Under Python 3 it does nothing.

    To support Python 2 and 3 with a single code base, define a __str__ method
    returning text and apply this decorator to the class.
    """
    if six.PY2:
        if '__str__' not in klass.__dict__:
            raise ValueError("@python_2_unicode_compatible cannot be applied "
                             "to %s because it doesn't define __str__()." %
                             klass.__name__)
        klass.__unicode__ = klass.__str__
        klass.__str__ = lambda self: self.__unicode__().encode('utf-8')
    return klass
